from __future__ import annotations

from typing import (Any, Callable, Dict, Generic, List, Literal, Optional,
                    Tuple, Type, TypeVar, Union, cast)
from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from .json import dumps, loads

Vector = List[float]
Value = Union[str, int, float, bool, List[str]]
MetaData = Dict[str, Value]
Filter = Literal["$eq", "$ne", "$lt", "$lte", "$gt", "$gte", "$in", "$nin"]
AndOr = Literal["$and", "$or"]
Query = Union[
    Dict[str, Union[Value, "Query", List[Value], List["Query"]]],
    Dict[AndOr, List[Dict[str, Union[Value, "Query", List[Value], List["Query"]]]]],
]


class CheapModel(BaseModel):
    def json(self, *args, **kwargs) -> str:
        return dumps(self.dict())

    @classmethod
    def parse(cls, s: str) -> CheapModel:
        return cls(**loads(s))


class QueryBuilder:
    def __init__(self, field: str = None, query: Query = None):  # type: ignore
        self.field = field
        self.query = query if query else {}

    def __repr__(self) -> str:
        return f"{self.query}"

    def __call__(self, field_name: str) -> QueryBuilder:
        return QueryBuilder(field_name)

    def __and__(self, other: QueryBuilder) -> QueryBuilder:
        return QueryBuilder(query={"$and": [self.query, other.query]})

    def __or__(self, other: QueryBuilder) -> QueryBuilder:
        return QueryBuilder(query={"$or": [self.query, other.query]})

    def __eq__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$eq": value}})

    def __ne__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$ne": value}})

    def __lt__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$lt": value}})

    def __le__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$lte": value}})

    def __gt__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$gt": value}})

    def __ge__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$gte": value}})

    def in_(self, values: List[Value]) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$in": values}})

    def nin_(self, values: List[Value]) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$nin": values}})



from __future__ import annotations

import asyncio
from asyncio import Lock
from dataclasses import dataclass, field
from os import environ
from typing import Dict, List, Optional, Type

from aiohttp import ClientSession
from aiohttp.client import ClientResponse, ClientTimeout, TCPConnector
from dotenv import load_dotenv
from multidict import CIMultiDict
from pydantic import BaseModel, Field

from .proxy import LazyProxy
from .schemas import (Embedding, MetaData, Query, QueryMatch, QueryRequest,
                      QueryResponse, UpsertRequest, UpsertResponse, Vector)
from .typedefs import Filter, QueryBuilder, Value

load_dotenv()



@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class APIClient(LazyProxy[ClientSession]):
    """
    HTTP Client:
    Base class to create HTTP clients that wrap `aiohttp.ClientSession`.
    Provides Lazy Loading through proxy objects and session reuse using the singleton pattern.
    Constructor Signature:
    `base_url` (str): Base URL of the API. For example: `https://api.openai.com`. Must the an absolute URL.
    `headers` (dict): Headers that will proxide the authentication credentials, content_type, etc.
    """

    base_url: str = field(init=True, repr=True)
    headers: Dict[str, str] = field(default_factory=dict)
    _subclasses: Optional[List[Type[APIClient]]] = None
    _session_creation_lock = asyncio.Lock()
    _session: Optional[ClientSession] = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._subclasses is None:
            cls._subclasses = []
        cls._subclasses.append(cls)

    @classmethod
    async def cleanup(cls):
        if hasattr(cls, "_subclasses") and cls._subclasses is not None:
            tasks = []
            for subclass in cls._subclasses:
                if subclass._session is not None:
                    tasks.append(subclass._session.close())
            await asyncio.gather(*tasks)
            cls._subclasses = None

    async def __load__(self) -> ClientSession:
        async with self._session_creation_lock:
            if self._session is None:
                self._session = ClientSession(
                    base_url=self.base_url,
                    headers=CIMultiDict(self.headers),
                    response_class=ClientResponse,
                    connector=TCPConnector(
                        keepalive_timeout=60,
                        ssl=False,
                        limit=1000,
                    ),
                    timeout=ClientTimeout(total=60),
                    connector_owner=False,
                    trust_env=False,
                    read_bufsize=2**18,
                )
        return self._session

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.base_url}>"


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class PineconeClient(APIClient):
    """
    Cheapcone went greedy so I had to make this one.
    """

    base_url: str = field(default=environ["PINECONE_API_URL"], init=True, repr=True)
    api_key: str = field(default=environ["PINECONE_API_KEY"], init=True, repr=False)

    def __load__(self) -> ClientSession:
        return ClientSession(base_url=self.base_url, headers={"api-key": self.api_key})

    async def upsert(self, embeddings:List[Embedding]) -> UpsertResponse:
        async with self.__load__() as session:
            values = []
            metadata = []
            for embedding in embeddings:
                values.append(embedding.values)
                metadata.append(embedding.metadata)
            
            async with session.post(
                "/vectors/upsert", json={"vectors": [UpsertRequest(values=values, metadata=metadata).dict() for values, metadata in zip(values, metadata)]}
            ) as response:
                return UpsertResponse(**await response.json())

    async def query(
        self, expr: Query, vector: Vector, includeMetadata: bool = True, topK: int = 10
    ) -> QueryResponse:
        async with self.__load__() as session:
            payload = QueryRequest(
                    topK=topK,
                    filter=expr,
                    vector=vector,
                    includeMetadata=includeMetadata,
                ).dict()
            async with session.post(
                "/query",
                json=payload,
            ) as response:
                return QueryResponse(**await response.json())


from datetime import datetime
from json import JSONDecoder, JSONEncoder
from typing import Any, NamedTuple
from uuid import UUID

from pydantic import BaseModel


class PineconeEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, datetime):
            return obj.astimezone().isoformat()
        elif isinstance(obj, NamedTuple):
            return obj._asdict()
        elif isinstance(obj, BaseModel):
            return obj.dict()
        else:
            return super().default(obj)


class PineconeDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        kwargs["object_hook"] = self.object_hook
        super().__init__(*args, **kwargs)

    def object_hook(self, obj: Any) -> Any:
        if "uuid" in obj:
            return UUID(obj["uuid"])
        elif "datetime" in obj:
            return datetime.fromisoformat(obj["datetime"])
        elif "type" in obj and "fields" in obj:
            return self.named_tuple_hook(obj)
        elif "type" in obj and "data" in obj:
            return self.pydantic_hook(obj)
        else:
            return obj

    def named_tuple_hook(self, obj: Any) -> Any:
        cls = NamedTuple(obj["type"], obj["fields"])
        return cls(*obj["values"])

    def pydantic_hook(self, obj: Any) -> Any:
        cls = getattr(__import__(obj["type"]), obj["type"])
        return cls(**obj["data"])


def dumps(obj: Any) -> str:
    return PineconeEncoder().encode(obj)


def loads(s: str) -> Any:
    return PineconeDecoder().decode(s)
